@@ -1,10 +1,11 @@
-from datetime import datetime
+from datetime import datetime, timedelta
 from inspect import stack
 from logging import Logger, getLogger
 
 from enlighten import Manager, Counter
 from numpy import where, ndarray, NaN
-from pandas import DataFrame, MultiIndex, NaT
+from pandas import DataFrame, MultiIndex, NaT, Series
+from scipy.special import psi
 
 from com.enovation.helper.pandas_dataframe_typer import PandasDataframeTyper
 import enlighten
@@ -49,14 +50,59 @@ class DatePredictabilityComputer:
 
     # Columns labels for working columns
     str__output__column_label__predictability: str = "predictability"
+    str__output__column_label__duration: str = "duration"
 
     # Default resampling frequency: 1D
     _str__default_resampling_rule: str = "1D"
 
-    def compute_date_predictability(
+    def compute_historical_date_predictability(
             self,
             df_measures: DataFrame,
+            b_resample: bool = True
     ) -> PredictabilityBean:
+        """
+        Function to compute date predictability for a set of measures provided as parameter.
+
+        The parameter df_measures should comply to the below:
+        - Instance of pandas.DataFrame
+        - BR_Structure_001: the measures do not have multiindex index
+        - BR_Structure_002: the measures do not have multiindex columns
+        - BR_Structure_003: the measures contains exactly 3 columns:
+            - str__input__column_label__key: str = "key"
+            - str__input__column_label__date: str = "date"
+            - str__input__column_label__measure: str = "measure"
+        - BR_Data_001: at least 1 record is required in order to compute a predictability
+        - BR_Data_002: there should not be any null value in the dataframe
+        - BR_Data_003: both columns "date" and "measure" should be dates, or objects that can be typed to dates
+        - BR_Data_004: there should not be any line with duplicated "date"
+
+        Depending on the parameter b_resample, the logic differs:
+        - If b_resample == True (default):
+            - The logic takes more time, but enable to graph precisely the historical trend of the date_predictability
+            - The measures are resampled daily. Illustration:
+                - For an input dataframe that contains 2 rows, measure 'x' for date '1-Jan' and measure 'y' for date
+                  '3-Jan'
+                - We resample daily, leading to dataframe that contains 3 rows, measure 'x' for dates '1-Jan' and
+                  '2-Jan' and measure 'y' for date '3-Jan'
+            - The date predictability is then computed as
+                - SUM for_each_measure(spot_predictability)
+                - With spot_predictability =
+                    - 0 if measure beyond the convergence_cone
+                    - otherwise (measure - last_measure)/convergence_cone
+                - With convergence_cone defined by:
+                    - Upper limit: date + abs(date - last_measure)
+                    - Lower limit: date - abs(date - last_measure)
+        - If b_resample == True:
+            - The logic is more optimal than when sampling, but graphing the historical trend of the
+              date_predictability is far less meaningful, and even misleading
+            - The date predictability is computed as:
+                - For each measure that spans from its date until the next one, we compute date_predictability as
+                    - SUM for_each_day (spot_predictabilty) with spot_predictability as defined above
+
+        :param df_measures: a dataframe that contains the measures
+        :param b_resample: whether or not we want to resample daily the measures, for later graphing
+        :return: the date predictability for the measures provided as a parameter
+        """
         self._logger.debug(
             f"Function '{stack()[0].filename} - {stack()[0].function}' is called with parameters:"
             f"\n\t- df_measures: '{df_measures.shape}'"
@@ -71,43 +117,41 @@ class DatePredictabilityComputer:
 
         # 2. We check and type the data
         the_df_typed = self._check_and_cleanse_data(
-            df_measures=df_measures,
+            df_measures=df_measures.copy(),
             str_column_label_key=self.str__input__column_label__key,
             str_column_label_date=self.str__input__column_label__date,
             str_column_label_measure=self.str__input__column_label__measure,
         )
-        self._logger.info(f"Function '{stack()[0].filename} - {stack()[0].function}': date checked and typed.")
 
-        # 3. We resample
-        the_df_resampled: DataFrame = self._resample(
-            df_measures=the_df_typed,
-            str_column_label_key=self.str__input__column_label__key,
-            str_column_label_date=self.str__input__column_label__date,
-            str_resample_rule=self._str__default_resampling_rule,
-        )
-        self._logger.info(f"Function '{stack()[0].filename} - {stack()[0].function}': measures resampled.")
+        # 3. Depending on whether or not we need to resample, we call the function to compute historical predictability
+        if b_resample:
+            df_the_historical_predictability: DataFrame = self._resample_and_compute_historical_date_predictability(
+                df_measures=the_df_typed.copy(),
+                str_column_label_key=self.str__input__column_label__key,
+                str_column_label_date=self.str__input__column_label__date,
+                str_column_label__measure=self.str__input__column_label__measure,
+                str_column_label__predictability=self.str__output__column_label__predictability,
+                str_resample_rule=self._str__default_resampling_rule,
+            )
+        else:
+            df_the_historical_predictability: DataFrame = \
+                self._compute_historical_date_predictability_without_resampling(
+                    df_measures=the_df_typed.copy(),
+                    str_column_label__key=self.str__input__column_label__key,
+                    str_column_label__date=self.str__input__column_label__date,
+                    str_column_label__measure=self.str__input__column_label__measure,
+                    str_column_label__duration=self.str__output__column_label__duration,
+                    str_column_label__predictability=self.str__output__column_label__predictability,
+                )
 
-        # 4. We compute historical predictability
-        the_df_historical_predictability: DataFrame = self._compute_historical_predictability(
-            df_measures=the_df_resampled,
-            str_column_label__key=self.str__input__column_label__key,
-            str_column_label__date=self.str__input__column_label__date,
-            str_column_label__measure=self.str__input__column_label__measure,
-            str_column_label__predictability=self.str__output__column_label__predictability
-        )
-        self._logger.info(f"Function '{stack()[0].filename} - {stack()[0].function}': predictability computed.")
-
-        # 5. We instantiate return bean
+        # 4. We compute the statistics
         df_the_statistics_by_key: DataFrame = self._compute_statistics_by_key(
-            df_measures=the_df_typed,
-            df_historical_predictability=the_df_historical_predictability,
-
+            df_measures=the_df_typed.copy(),
+            df_historical_predictability=df_the_historical_predictability,
             str_column_label__key=self.str__input__column_label__key,
             str_column_label__date=self.str__input__column_label__date,
             str_column_label__measure=self.str__input__column_label__measure,
-
             str_column_label__predictability=self.str__output__column_label__predictability,
-
             str_column_label__measure_count=self.str__statistics__column_label__measure_count,
             str_column_label__date_first=self.str__statistics__column_label__date_first,
             str_column_label__date_last=self.str__statistics__column_label__date_last,
@@ -121,7 +165,7 @@ class DatePredictabilityComputer:
         self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")
 
         return PredictabilityBean(
-            df_historical=the_df_historical_predictability,
+            df_historical=df_the_historical_predictability,
             df_by_measure=df_measures,
             df_by_key=df_the_statistics_by_key,
         )
@@ -227,7 +271,7 @@ class DatePredictabilityComputer:
         # LOG_Data_001: Columns "date" and "measure" are typed into dates
         try:
             df_the_return: DataFrame = PandasDataframeTyper.type(
-                df_to_type=df_measures.copy(),
+                df_to_type=df_measures,
                 dict_columns_to_type={
                     str_column_label_date: "date",
                     str_column_label_measure: "date"
@@ -256,21 +300,21 @@ class DatePredictabilityComputer:
                 str_column_label_date
             ]
         ).reset_index(drop=True)
-        df_the_return["tmp_"+str_column_label_key] = df_the_return[str_column_label_key].shift(1)
-        df_the_return["tmp_"+str_column_label_measure] = df_the_return[str_column_label_measure].shift(1)
+        df_the_return["tmp_" + str_column_label_key] = df_the_return[str_column_label_key].shift(1)
+        df_the_return["tmp_" + str_column_label_measure] = df_the_return[str_column_label_measure].shift(1)
 
         # We override the very last measure for each key, not to drop it in case it is similar to previous one
         # df_the_return = df_the_return.loc[-1, "tmp_"+str_column_label_key] = NaN
         # df_the_return.loc[-1, "tmp_"+str_column_label_measure] = NaT
         df_the_return.loc[
             df_the_return.groupby(str_column_label_key).tail(1).index,
-            ["tmp_"+str_column_label_key, "tmp_"+str_column_label_measure]
+            ["tmp_" + str_column_label_key, "tmp_" + str_column_label_measure]
         ] = [NaN, NaT]
 
         df_the_return = df_the_return[
-            (df_the_return[str_column_label_key] != df_the_return["tmp_"+str_column_label_key])
-            | (df_the_return[str_column_label_measure] != df_the_return["tmp_"+str_column_label_measure])
-        ]
+            (df_the_return[str_column_label_key] != df_the_return["tmp_" + str_column_label_key])
+            | (df_the_return[str_column_label_measure] != df_the_return["tmp_" + str_column_label_measure])
+            ]
         df_the_return = df_the_return.drop(columns=[
             "tmp_" + str_column_label_key,
             "tmp_" + str_column_label_measure
@@ -280,7 +324,40 @@ class DatePredictabilityComputer:
 
         return df_the_return
 
-    def _resample(
+    def _resample_and_compute_historical_date_predictability(
+            self,
+            df_measures: DataFrame,
+            str_column_label_key: str = str__input__column_label__key,
+            str_column_label_date: str = str__input__column_label__date,
+            str_column_label__measure: str = str__input__column_label__measure,
+            str_column_label__predictability: str = str__output__column_label__predictability,
+            str_resample_rule: str = _str__default_resampling_rule,
+    ) -> DataFrame:
+
+        # 1. We resample
+        the_df_resampled: DataFrame = self._resample_measures(
+            df_measures=df_measures,
+            str_column_label_key=str_column_label_key,
+            str_column_label_date=str_column_label_date,
+            str_resample_rule=str_resample_rule,
+        )
+        self._logger.info(f"Function '{stack()[0].filename} - {stack()[0].function}': measures resampled.")
+
+        # 2. We compute historical predictability
+        the_df_historical_predictability: DataFrame = \
+            self._compute_historical_date_predictability_for_resampled_measures(
+                df_measures=the_df_resampled,
+                str_column_label__key=str_column_label_key,
+                str_column_label__date=str_column_label_date,
+                str_column_label__measure=str_column_label__measure,
+                str_column_label__predictability=str_column_label__predictability
+            )
+
+        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")
+
+        return the_df_historical_predictability
+
+    def _resample_measures(
             self,
             df_measures: DataFrame,
             str_column_label_key: str = str__input__column_label__key,
@@ -290,70 +367,19 @@ class DatePredictabilityComputer:
 
         self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called")
 
-        df_measures = df_measures\
-            .set_index(str_column_label_date)\
-            .groupby(str_column_label_key)\
-            .resample(str_resample_rule)\
-            .ffill()\
-            .reset_index(level=str_column_label_date)\
+        df_measures = df_measures \
+            .set_index(str_column_label_date) \
+            .groupby(str_column_label_key) \
+            .resample(str_resample_rule) \
+            .ffill() \
+            .reset_index(level=str_column_label_date) \
             .reset_index(drop=True)
 
         self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning")
 
         return df_measures
 
-    def _compute_spot_predictability_for_a_given_key(
-            self,
-            df_measures: DataFrame,
-            str_column_label__date: str = str__input__column_label__date,
-            str_column_label__measure: str = str__input__column_label__measure,
-    ) -> float:
-        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")
-
-        dt_the_last_date: datetime = df_measures[self.str__input__column_label__date].max()
-
-        dt_the_last_measure: datetime = \
-            df_measures[
-                df_measures[self.str__input__column_label__date] == dt_the_last_date
-            ][
-                self.str__input__column_label__measure
-            ].iloc[0]
-
-        a_measures_predictabilities: ndarray = where(
-
-            # If date is at last measure
-            df_measures[str_column_label__date] == dt_the_last_measure,
-
-            where(
-
-                # If measure is spot on
-                df_measures[str_column_label__measure] == dt_the_last_measure,
-                1,
-                0
-            ),
-
-            where(
-
-                # If outside the cone
-                abs((df_measures[str_column_label__measure]-dt_the_last_measure).dt.days)
-                >= abs((df_measures[str_column_label__date]-dt_the_last_measure).dt.days),
-                0,
-                1
-                - abs((df_measures[str_column_label__measure] - dt_the_last_measure).dt.days)
-                / abs((df_measures[str_column_label__date] - dt_the_last_measure).dt.days)
-            )
-        )
-
-        # Eventually, we compute the aggregated predictability
-        f_the_return: float = \
-            a_measures_predictabilities.sum() \
-            / len(a_measures_predictabilities)
-
-        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
-
-        return f_the_return
-
-    def _compute_historical_predictability(
+    def _compute_historical_date_predictability_for_resampled_measures(
             self,
             df_measures: DataFrame,
             str_column_label__key: str = str__input__column_label__key,
@@ -399,15 +425,14 @@ class DatePredictabilityComputer:
 
         # For each line in the dataframe
         for t_i_row in df_measures.itertuples():
-
             # We compute the predictability:
             # - For the given key
             # - For all the previous records at the date
-            f_i_predictability = self._compute_spot_predictability_for_a_given_key(
+            f_i_predictability = self._compute_date_predictability_for_a_given_key_and_resampled_measures(
                 df_measures=df_measures[
                     (df_measures[str_column_label__key] == getattr(t_i_row, str_column_label__key))
                     & (df_measures[str_column_label__date] <= getattr(t_i_row, str_column_label__date))
-                ],
+                    ],
                 str_column_label__date=str_column_label__date,
                 str_column_label__measure=str_column_label__measure,
             )
@@ -422,6 +447,213 @@ class DatePredictabilityComputer:
 
         return df_measures
 
+    def _compute_date_predictability_for_a_given_key_and_resampled_measures(
+            self,
+            df_measures: DataFrame,
+            str_column_label__date: str = str__input__column_label__date,
+            str_column_label__measure: str = str__input__column_label__measure,
+    ) -> float:
+        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")
+
+        dt_the_last_date: datetime = df_measures[self.str__input__column_label__date].max()
+
+        dt_the_last_measure: datetime = \
+            df_measures[
+                df_measures[self.str__input__column_label__date] == dt_the_last_date
+                ][
+                self.str__input__column_label__measure
+            ].iloc[0]
+
+        a_measures_predictabilities: ndarray = where(
+
+            # If date is at last measure
+            df_measures[str_column_label__date] == dt_the_last_measure,
+
+            where(
+
+                # If measure is spot on
+                df_measures[str_column_label__measure] == dt_the_last_measure,
+                1,
+                0
+            ),
+
+            where(
+
+                # If outside the cone
+                abs((df_measures[str_column_label__measure] - dt_the_last_measure).dt.days)
+                >= abs((df_measures[str_column_label__date] - dt_the_last_measure).dt.days),
+                0,
+                1
+                - abs((df_measures[str_column_label__measure] - dt_the_last_measure).dt.days)
+                / abs((df_measures[str_column_label__date] - dt_the_last_measure).dt.days)
+            )
+        )
+
+        # The last measure should not contribute tot he aggregated predictability
+        # Therefore, we remove the last record from the ndarray (only if more than 1 records)
+        if len(a_measures_predictabilities) > 1:
+            a_measures_predictabilities = a_measures_predictabilities[:-1]
+
+        # Eventually, we compute the aggregated predictability
+        f_the_return: float = \
+            a_measures_predictabilities.sum() \
+            / len(a_measures_predictabilities)
+
+        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
+
+        return f_the_return
+
+    def _compute_historical_date_predictability_without_resampling(
+            self,
+            df_measures: DataFrame,
+            str_column_label__key: str = str__input__column_label__key,
+            str_column_label__date: str = str__input__column_label__date,
+            str_column_label__measure: str = str__input__column_label__measure,
+            str_column_label__predictability: str = str__output__column_label__predictability,
+            str_column_label__duration: str = str__output__column_label__duration,
+    ) -> DataFrame:
+        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")
+
+        the_df_enriched_measures: DataFrame = df_measures
+
+        # Setup progress bar
+        obj_the_mgr: Manager = enlighten.get_manager()
+        obj_the_counter: Counter = obj_the_mgr.counter(
+            total=len(df_measures.index),
+            desc='Progress:',
+            unit='measures'
+        )
+
+        # We instantiate the column to bear the predictability
+        the_df_enriched_measures[str_column_label__predictability] = -1
+
+        # For each line in the dataframe
+        for t_i_row in df_measures.itertuples():
+            # We compute the predictability:
+            # - For the given key
+            # - For all the previous records at the date
+            f_i_predictability = self._compute_date_predictability_for_a_given_key_not_resampled(
+                df_measures=the_df_enriched_measures[
+                    (the_df_enriched_measures[str_column_label__key] == getattr(t_i_row, str_column_label__key))
+                    & (the_df_enriched_measures[str_column_label__date] <= getattr(t_i_row, str_column_label__date))
+                    ],
+                str_column_label__date=str_column_label__date,
+                str_column_label__measure=str_column_label__measure,
+                str_column_label__duration=str_column_label__duration,
+            )
+
+            # We update the measures with these historical predictabilities
+            df_measures.loc[getattr(t_i_row, 'Index'), str_column_label__predictability] = f_i_predictability
+
+            # We update the progress bar
+            obj_the_counter.update()
+
+        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
+
+        return df_measures
+
+    def _compute_date_predictability_for_a_given_key_not_resampled(
+            self,
+            df_measures: DataFrame,
+            str_column_label__date: str = str__input__column_label__date,
+            str_column_label__measure: str = str__input__column_label__measure,
+            str_column_label__duration: str = str__output__column_label__duration,
+    ) -> float:
+        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is called.")
+
+        the_df_enriched_measures: DataFrame = df_measures.copy()
+
+        # We duplicate column "date", that we shift by one
+        # Assumption: values already sorted in function _check_and_cleanse_data
+        the_df_enriched_measures["date_end"] = the_df_enriched_measures[str_column_label__date].shift(-1)
+        the_df_enriched_measures.loc[the_df_enriched_measures.index[-1], "date_end"] = \
+            the_df_enriched_measures.loc[the_df_enriched_measures.index[-1],
+                                         DatePredictabilityComputer.str__input__column_label__date]
+
+        # We compute the duration for each measure
+        the_df_enriched_measures[str_column_label__duration] = (
+                the_df_enriched_measures["date_end"] - the_df_enriched_measures[str_column_label__date]).dt.days
+
+        dt_the_last_date: datetime = df_measures[self.str__input__column_label__date].max()
+        dt_the_last_measure: datetime = df_measures[
+            df_measures[self.str__input__column_label__date] == dt_the_last_date
+            ][
+            self.str__input__column_label__measure
+        ].iloc[0]
+
+        def compute_magic_number(s_row: Series) -> float:
+            dt_the_date_start: datetime = s_row[str_column_label__date]
+            dt_the_date_end: datetime = s_row["date_end"]
+            dt_the_measure: datetime = s_row[str_column_label__measure]
+            int_the_duration: int = s_row[str_column_label__duration]
+
+            # For the very last measure, the closing one,
+            if int_the_duration == 0:
+                return 0.0
+
+            # If we are spot on
+            if dt_the_measure == dt_the_last_measure:
+                return int_the_duration
+
+            # If date_start is before last measure, and within convergence cone
+            if (dt_the_date_start < dt_the_last_measure) & (
+                    abs((dt_the_measure - dt_the_last_measure).days)
+                    < abs((dt_the_last_measure - dt_the_date_start).days)
+            ):
+                dt_the_effective_date_end: datetime = min(
+                    dt_the_date_end,
+                    dt_the_last_measure-timedelta(days=abs((dt_the_measure - dt_the_last_measure).days))
+                )
+                # int_the_cone_start: int = abs((dt_the_last_measure-dt_the_date_start).days)
+                int_the_cone_effective_end: int = abs((dt_the_last_measure-dt_the_effective_date_end).days)+1
+
+                # We prepare the operands
+                int_b: int = int_the_cone_effective_end
+                int_a: int = int_the_cone_effective_end - abs((dt_the_last_measure-dt_the_measure).days)
+                int_m: int = (dt_the_effective_date_end-dt_the_date_start).days
+
+                f_the_predictability: float = ((int_b-int_a)*(psi(int_b)-psi(int_b+int_m))+int_m)
+            else:
+                f_the_predictability: float = 0
+
+            # If date_end is after last measure, and within convergence cone
+            if (dt_the_date_end > dt_the_last_measure) & (
+                    abs((dt_the_measure - dt_the_last_measure).days)
+                    < abs((dt_the_last_measure - dt_the_date_end).days)
+            ):
+                dt_the_effective_date_start: datetime = max(
+                    dt_the_date_start,
+                    dt_the_last_measure+timedelta(days=abs((dt_the_measure - dt_the_last_measure).days))
+                )
+                # int_the_cone_end: int = abs((dt_the_last_measure-dt_the_date_end).days)-1
+                int_the_cone_effective_start: int = abs((dt_the_last_measure-dt_the_effective_date_start).days)
+
+                # We prepare the operands
+                int_b: int = int_the_cone_effective_start
+                int_a: int = int_the_cone_effective_start - abs((dt_the_last_measure-dt_the_measure).days)
+                int_m: int = (dt_the_date_end-dt_the_effective_date_start).days
+
+                f_the_predictability += ((int_b-int_a)*(psi(int_b)-psi(int_b+int_m))+int_m)
+            else:
+                pass
+
+            return f_the_predictability
+
+        if len(df_measures.index) == 1:
+            f_the_return: float = 1.0
+
+        else:
+            the_s_historical_predictability: Series = the_df_enriched_measures.apply(compute_magic_number, axis=1)
+
+            # Eventually, we compute the aggregated predictability
+            f_the_return: float = \
+                the_s_historical_predictability.sum() / \
+                the_df_enriched_measures[str_column_label__duration].sum()
+
+        self._logger.debug(f"Function '{stack()[0].filename} - {stack()[0].function}' is returning.")
+
+        return f_the_return
+
     def _compute_statistics_by_key(
             self,
             df_measures: DataFrame,
